/*--------------------------------------------------------------
// file: bigG.c - compute Gaussians on each pairwise vector 
//   G = bigG(A,B,sigmaSqr)
// INPUTS:
//   A - mxd matrix (m=number of points, d=number of dimensions)
//   B - kxd matrix (k=number of points, d=number of dimensions)
//   sigmaSqr - real scalar representing square of the standard deviation
// OUTPUTS:
//   G - mxk matrix of evaluations of a multivariate normal 
//       with mean zero and covariance = sigmaSqr I for each 
//       pair of inputs, i.e. 
//       G_{i,j} = mvnpdf( A(i,:)-B(j,:), 0, sigmaSqr I ); 
//
// Checked: 
//--------------------------------------------------------------*/
#include <math.h>
#include "mex.h"

/*--------------------------------------------------------------
// function: mexFunction - Entry point from Matlab environment
// INPUTS:
//   nlhs - number of left hand side arguments (outputs)
//   plhs[] - pointer to table where created matrix pointers are
//            to be placed
//   nrhs - number of right hand side arguments (inputs)
//   prhs[] - pointer to table of input matrices
//--------------------------------------------------------------*/
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
  int M,N,K; /* Sizes */
  int m,n,k; /* Loop variables */
  double *pA, *pB, *pS, *pG, sigmaSqr, denom;

  if (nrhs < 3)
    mexErrMsgTxt("Not enough input arguments.");
  if (nrhs > 3)
    mexErrMsgTxt("Too many input arguments.");
  if (!mxIsDouble(prhs[0]) || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]))
    mexErrMsgTxt("Input arguments must be type double.");
  if (mxGetN(prhs[0]) != mxGetN(prhs[1]))
    mexErrMsgTxt("Input matrix dimensions must agree (number of columns).");
  if (mxIsComplex(prhs[0]) || mxIsComplex(prhs[1]) || mxIsComplex(prhs[2]))
    mexWarnMsgTxt("Complex parts ignored.");

  /* bigG(MxK,NxK) => MxN */
  M = mxGetM(prhs[0]); /* Number of rows in output */
  N = mxGetM(prhs[1]); /* Number of columns in output */
  K = mxGetN(prhs[0]); /* Number of dimensions */

  plhs[0] = mxCreateDoubleMatrix(M,N,mxREAL);

  pA = mxGetPr(prhs[0]);
  pB = mxGetPr(prhs[1]);
  pS = mxGetPr(prhs[2]); sigmaSqr = pS[0];
  pG = mxGetPr(plhs[0]);

  denom = pow( 2.*M_PI*sigmaSqr, K/2. );

  /* Now the distance computation */
  for (m=0; m<M; m++)
  {
    for (n=0; n<N; n++)
    {
      double sum = 0.0, diff;
      for (k=0; k<K; k++)
      {
        diff = pA[m+M*k] - pB[n+N*k];
        sum += diff*diff;
      }
      pG[m+M*n] = exp( -0.5*sum/sigmaSqr )/denom;
    }
  } 
} /* end mexFunction() */
