function [ Ys, W, I ] = MMI_SG(Xs,k,Jp,d,W0)
%
% Maximize mutual information between projected space and the class label
%
% Inputs:
%   Xs: original feature patterns of size [ D, N ]
%   k: pattern labels of size [ N, 1 ]
%   Jp: 
%   d: desired reduced dimension (must satisfy 1 <= d <= D)
%   W0: initial guess at the projection matrix of size [ D, d ]
% Outputs:
%   W: final projection matrix of size [ D, d ]
% 
% Written by:
% -- 
% John L. Weatherwax                2005-10-30
% 
% email: wax@alum.mit.edu
% 
% Please send comments and especially bug reports to the
% above email address.
% 
%-----

MAX_OUTER_LOOPS=10; 
MAX_INNER_LOOPS=5; 
MIN_INCREASE_IN_MI=10^-10; %10^(-4); 

%
% Code begins:

[ N, D ] = size(Xs); 
if( d>D ) error( 'Input value of d is too large' ); end

W = W0; 
Ys = Xs * W;
sgma = halfTwoFarthestPoints( Ys ); %sgma=5;
no=0; 
fprintf('>s: no=%3d: sgma=%g\n',no,sgma);
while( 1 ) % sgma reduction/refinement loop
  ni=0; 
  while( 1 ) % mutual information reduction loop
    
    % Randomly select two samples from different classes (according to their priors):
    Pc = Jp/N; 
    c1 = sample_discrete( Pc, 1, 1 ); 
    Pc(c1) = 0; Pc = Pc/sum(Pc); 
    c2 = sample_discrete( Pc, 1, 1 ); 
    fprintf( '  Peterb c1=%3d; c2=%3d\n', c1, c2 ); 
    inds1 = find(k==c1); inds2 = find(k==c2);
    sInds1 = shuffle( inds1 ); i1 = sInds1(1); 
    sInds2 = shuffle( inds2 ); i2 = sInds2(1); 
    
    V_in  = 0.5*mvnpdf(zeros(d,1),zeros(d,1),2*eye(d)*sgma^2);
    V_all = 0.25*mvnpdf(zeros(d,1),zeros(d,1),2*eye(d)*sgma^2) + ...
	    0.25*mvnpdf( (Ys(i1,:)-Ys(i2,:)).', zeros(d,1),2*eye(d)*sgma^2);
    V_btw = V_all; 
    I_T = V_in + V_all - 2*V_btw; I = I_T;
  
    dI_TdY1 = (-1/(8*(sgma^2)))*mvnpdf( (Ys(i1,:)-Ys(i2,:)).', zeros(d,1),2*eye(d)*sgma^2) * ...
	(Ys(i1,:)-Ys(i2,:)).';
  
    Wud = (-1/(8*(sgma^2)))*mvnpdf( (Ys(i1,:)-Ys(i2,:)).', zeros(d,1),2*eye(d)*sgma^2) * ...
        ( (Ys(i2,:)-Ys(i1,:)).' ) * ( Xs(i1,:) - Xs(i2,:) ); Wud = Wud.'; 

    % Plot this projection: 
    if( 0 ) 
      cs = { 'yo', 'g*', 'bd', 'rs', 'bx', 'm+' }; 
      figure; 
      for i=1:length(unique(k))
	inds=find(k==i);
	plot( -Ys(inds,1), +Ys(inds,2), cs{i} ); 
	%plot( +Ys(inds,1), -Ys(inds,2), cs{i} ); 
	%plot( -Ys(inds,1), -Ys(inds,2), cs{i} ); 
	if( i==1 ) hold on; end
      end
      axis( [ -3.5 3.5 -3.5 3.5 ] ); grid on; 
      %title( ['Projected Data; I=',num2str(I)] ); drawnow; 
      title( ['Projected Data' ] ); drawnow; 
    end
 
    % Perform very naive line search to find an "optimal" eta:
    % --first bracket our maximum by testing a great number of hypothesis:
    
    % Begin a more refined bracket:
    topExp = ceil( -log10(norm(Wud)) )+1; 
% $$$     [br_min, br_mid, br_max, num_evals] = minbrack( 'miMinFn', 0, 10^topExp, -I, ...
% $$$ 	                                            W, Wud, Xs, k, Jp, sgma ); 
% $$$     fprintf( '  Brack found [%6g, %6g, %6g]; in %3d evals\n', br_min, br_mid, br_max, num_evals ); 
    %etaRng = 10.^[ -1:topExp ]; 
    etaRng = 10.^[ -1:8 ]; 
    for ei=1:length(etaRng)
      Wp = W + etaRng(ei) * Wud; 
      Q = gram(Wp); Wp = Q; %clear Q; 
      Ys = Xs * Wp;
      [ VCY2, VC2Y2, VCY ] = computeInformationPotentials(Ys,k,Jp,sgma); 
      IT = VCY2 + VC2Y2 - 2*VCY;
      fprintf( '  %5d: ETA=%10g; IT=%20g', ei, etaRng(ei), IT ); 
      if( ei~=length(etaRng) ) fprintf('\n'); end
      Is(ei) = IT; 
    end
    % select the best eta: 
    [dum,ii]=max(Is); fprintf('          i_opt=%d\n',ii);
    etaS = etaRng(ii); 
    if( 0 ) 
    if( ii==1 ) 
      pt=etaRng(1); fpt=-Is(1); dir=-1;
    elseif( ii==length(etaRng) )
      pt=etaRng(end); fpt=-Is(end); dir=+1;
    else
      [br_min, br_mid, br_max, num_evals] = minbrack( 'miMinFn', etaRng(ii-1), etaRng(ii+1), -Is(ii-1), ...
                                                      W, Wud, Xs, k, Jp, sgma ); 
      fprintf( '  Brack found [%6g, %6g, %6g]; in %3d evals\n', br_min, br_mid, br_max, num_evals ); 
      pt=etaRng(ii-1); fpt=-Is(ii-1); dir=+1;
    end
    fprintf( '  Searching for maximum...\n' ); 
    options = foptions;
    [etaS, options] = linemin( 'miMinFn', pt, dir, fpt, options, W, Wud, Xs, k, Jp, sgma );
    fprintf( '  Found at %6g with value %6g\n', etaS, -miMinFn( etaS, W, Wud, Xs, k, Jp, sgma ) )
    end;
    W = W + etaS * Wud; 
    Q = gram(W); W = Q; %clear Q; 
    Ys = Xs * W;
   
    if( 0 ) 
      cs = { 'ko', 'k*', 'kd', 'ks', 'kx', 'k+' }; 
      for i=1:length(unique(k))
	inds=find(k==i);
	plot( -Ys(inds,1), Ys(inds,2), cs{i} ); 
      end
      axis( [ -3.5 3.5 -3.5 3.5 ] ); grid on; 
      title( 'Projected Data' );    
    end
  
    if( 0 ) 
      % Plot each iteration: 
      %cs = { 'yo', 'g*', 'bd', 'rs', 'bx', 'm+' }; 
      cs = { 'ko', 'g*', 'bd', 'rs', 'bx', 'm+' }; 
      figure; 
      for i=1:length(unique(k))
	inds=find(k==i);
	plot( -Ys(inds,1), Ys(inds,2), cs{i} ); 
	if( i==1 ) hold on; end
      end
      axis( [ -3.5 3.5 -3.5 3.5 ] ); grid on; drawnow;
    end
  
    %keyboard; 
    %pause; 

    if( ni==0 ) 
      deltaI = +inf; 
    else
      deltaI = I-I_old;
    end
    I_old=I; 
    fprintf( '  >f: ni=%3d: deltaI = %g\n', ni, deltaI ); 
    
    ni=ni+1; 
    if( ni > MAX_INNER_LOOPS | deltaI < MIN_INCREASE_IN_MI ) break; end
  end
  sgma=sgma/2; 
  fprintf( '>f: no=%3d: sgma=%g\n', no, sgma );
  no=no+1; 
  if( no > MAX_OUTER_LOOPS | sgma < halfInClassTwoFarthestPoints( Ys, k ) ) break; end
end
