/*--------------------------------------------------------------
// file: dMIdTheta.c - compute the derivative of the mutual information 
//       with respect to the vector of parameters theta
//   dMIdT = dMIdTheta(dIdY,Xs,dWdT)
// INPUTS:
//   dIdY - Nxd matrix (N=number of data points, d=reduced feature dimension size)
//   Xs   - NxD matrix (N=number of data points, D=full feature dimension size)
//   dWdT - Dx(d^2 (D-d)) block matrix.  We have d*(D-d) (the number of thetas) submatrices
//          (dWdTk) concatonated along the rows.  Each dWdTk submatrix is of size Dxd
// OUTPUTS:
//   dMIdT - column vector of length d*(D-d) each component represents 
//           dI/dTheta_k
//
// Checked: 2005-11-12 (WWX)
//
//  Written by:
//  -- 
//  John L. Weatherwax                2005-10-30
//  
//  email: wax@alum.mit.edu
//  
//  Please send comments and especially bug reports to the
//  above email address.
//  
//--------------------------------------------------------------*/
#include <math.h>
#include "mex.h"

/*--------------------------------------------------------------
// function: mexFunction - Entry point from Matlab environment
// INPUTS:
//   nlhs - number of left hand side arguments (outputs)
//   plhs[] - pointer to table where created matrix pointers are
//            to be placed
//   nrhs - number of right hand side arguments (inputs)
//   prhs[] - pointer to table of input matrices
//--------------------------------------------------------------*/
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
  int N,d,D; 
  int k,i,l,m,colLoc; /* Loop variables */
  double k_sum,il_sum,dIdyil,dyidWmn,dWmldTk; /* dummy variables */
  double *pdIdY,*pXs,*pdWdT,*pdMIdT;

  if (nrhs < 3)
    mexErrMsgTxt("Not enough input arguments.");
  if (nrhs > 3)
    mexErrMsgTxt("To many input arguments.");
  if (nlhs < 1)
    mexErrMsgTxt("Too few output arguments.");
  if (nlhs > 2)
    mexErrMsgTxt("Too many output arguments.");
  if (!mxIsDouble(prhs[0]) || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]))
    mexErrMsgTxt("Input arguments must be type double.");
  if (mxGetM(prhs[0]) != mxGetM(prhs[1]))
    mexErrMsgTxt("number of rows of argument 1 must equal that of argument 2.");
  if (mxGetN(prhs[1]) != mxGetM(prhs[2]))
    mexErrMsgTxt("number of columns of argument 2 must equal the number of rows of argument 3.");
  if (mxIsComplex(prhs[0]) || mxIsComplex(prhs[1]) || mxIsComplex(prhs[2]))
    mexWarnMsgTxt("Complex parts ignored.");

  /* dMIdTheta(Nxd,NxD,Dx(d^2 (D-d))) => d*(D-d) x 1 */
  N = mxGetM(prhs[0]); 
  d = mxGetN(prhs[0]); 
  D = mxGetN(prhs[1]); 

  /* Create output column vector */
  plhs[0] = mxCreateDoubleMatrix(d*(D-d),1,mxREAL);

  pdIdY  = mxGetPr(prhs[0]);
  pXs    = mxGetPr(prhs[1]);
  pdWdT  = mxGetPr(prhs[2]); 
  pdMIdT = (double *)mxGetData(plhs[0]);

  /* Now the computation */
  for( k=0; k<d*(D-d); k++ ){ /* Loop over each k theta derivative */ 
    k_sum=0; 

    for( i=0; i<N; i++ ){ /* Loop over each sample point */
      for( l=0; l<d; l++ ){ /* Loop over each of the lth components of the ith sample y */ 

        il_sum = 0; 
        dIdyil = pdIdY[i+N*l]; 
        for( m=0; m<D; m++ ){ /* Loop over each of the (m,n) components of W */ 
          colLoc  = k*d+l; /* ZERO based */ 
          dWmldTk = pdWdT[m + D*colLoc]; 
          il_sum += pXs[i+N*m] * dWmldTk; 
        } /* m */ 
        k_sum += dIdyil*il_sum;

      } /* l */ 
    } /* i */ 
    pdMIdT[k] = k_sum; 

  } /* k */ 
} /* end mexFunction() */
